@@ -20,6 +20,8 @@
 
 #include <dbus/dbus.h>
 #include <glib-object.h>
+#include <libhal.h>
+#include "oobs-session-private.h"
 #include "oobs-list-private.h"
 #include "oobs-object-private.h"
 #include "oobs-ifacesconfig.h"
@@ -29,7 +31,9 @@
 #include "oobs-iface-plip.h"
 #include "oobs-iface-modem.h"
 #include "oobs-iface-isdn.h"
+#include "iface-state-monitor.h"
 #include "utils.h"
+#include "config.h"
 
 #define IFACES_CONFIG_REMOTE_OBJECT "IfacesConfig"
 #define OOBS_IFACES_CONFIG_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), OOBS_TYPE_IFACES_CONFIG, OobsIfacesConfigPrivate))
@@ -47,6 +51,13 @@ struct _OobsIfacesConfigPrivate
 
   GList *available_config_methods;
   GList *available_key_types;
+
+  GHashTable *ifaces;
+
+#ifdef HAVE_HAL
+  LibHalContext *hal_context;
+  GHashTable    *devices;
+#endif
 };
 
 static void oobs_ifaces_config_class_init (OobsIfacesConfigClass *class);
@@ -97,6 +108,126 @@ oobs_ifaces_config_class_init (OobsIfacesConfigClass *class)
 			    sizeof (OobsIfacesConfigPrivate));
 }
 
+static void
+oobs_ifaces_config_iface_monitor (OobsIfacesConfig *config,
+				  const gchar      *iface_name,
+				  gboolean          iface_active)
+{
+  OobsIfacesConfigPrivate *priv;
+  OobsIface *iface;
+
+  priv = config->_priv;
+  iface = g_hash_table_lookup (priv->ifaces, iface_name);
+
+  if (!iface)
+    return;
+
+  g_return_if_fail (OOBS_IS_IFACE (iface));
+
+  if (iface_active != oobs_iface_get_active (iface))
+    {
+      oobs_iface_set_active (iface, iface_active);
+      g_signal_emit_by_name (iface, "state-changed");
+    }
+}
+
+#ifdef HAVE_HAL
+
+static void
+hal_context_device_added (LibHalContext *context,
+			  const gchar   *udi)
+{
+  OobsIfacesConfig *config;
+  OobsIfacesConfigPrivate *priv;
+
+  config = libhal_ctx_get_user_data (context);
+  priv = OOBS_IFACES_CONFIG_GET_PRIVATE (config);
+
+  g_return_if_fail (OOBS_IS_IFACES_CONFIG (config));
+
+  if (libhal_device_query_capability (context, udi, "net", NULL))
+    {
+      g_hash_table_insert (priv->devices, g_strdup (udi), GINT_TO_POINTER (TRUE));
+      g_signal_emit_by_name (config, "changed");
+    }
+}
+
+static void
+hal_context_device_removed (LibHalContext *context,
+			    const gchar   *udi)
+{
+  OobsIfacesConfig *config;
+  OobsIfacesConfigPrivate *priv;
+
+  config = libhal_ctx_get_user_data (context);
+  priv = OOBS_IFACES_CONFIG_GET_PRIVATE (config);
+
+  g_return_if_fail (OOBS_IS_IFACES_CONFIG (config));
+
+  if (g_hash_table_lookup (priv->devices, udi))
+    {
+      g_hash_table_remove (priv->devices, udi);
+      g_signal_emit_by_name (config, "changed");
+    }
+}
+
+static GHashTable*
+hal_context_get_initial_devices (LibHalContext *context)
+{
+  GHashTable *devices;
+  gint i, n_devices;
+  gchar **udis;
+
+  devices = g_hash_table_new_full (g_str_hash, g_str_equal,
+				   (GDestroyNotify) g_free, NULL);
+
+  udis = libhal_find_device_by_capability (context, "net", &n_devices, NULL);
+
+  for (i = 0; i < n_devices; i++)
+    g_hash_table_insert (devices, g_strdup (udis[i]), GINT_TO_POINTER (TRUE));
+
+  libhal_free_string_array (udis);
+
+  return devices;
+}
+
+static void
+init_hal_context (OobsIfacesConfig *config)
+{
+  OobsIfacesConfigPrivate *priv;
+  OobsSession *session;
+  DBusConnection *connection;
+  DBusError error;
+
+  priv = OOBS_IFACES_CONFIG_GET_PRIVATE (config);
+
+  dbus_error_init (&error);
+  session = oobs_session_get ();
+  connection = _oobs_session_get_connection_bus (session);
+
+  priv->hal_context = libhal_ctx_new ();
+  libhal_ctx_set_dbus_connection (priv->hal_context, connection);
+
+  libhal_ctx_set_user_data (priv->hal_context, config);
+
+  libhal_ctx_set_device_added (priv->hal_context,
+			       hal_context_device_added);
+  libhal_ctx_set_device_removed (priv->hal_context,
+			       hal_context_device_removed);
+
+  libhal_ctx_init (priv->hal_context, &error);
+
+  if (dbus_error_is_set (&error))
+    {
+      g_warning (error.message);
+      return;
+    }
+
+  priv->devices = hal_context_get_initial_devices (priv->hal_context);
+}
+
+#endif
+
 static void
 oobs_ifaces_config_init (OobsIfacesConfig *config)
 {
@@ -104,6 +235,10 @@ oobs_ifaces_config_init (OobsIfacesConfig *config)
 
   priv = OOBS_IFACES_CONFIG_GET_PRIVATE (config);
 
+#ifdef HAVE_HAL
+  init_hal_context (config);
+#endif
+
   priv->ethernet_ifaces = _oobs_list_new (OOBS_TYPE_IFACE_ETHERNET);
   priv->wireless_ifaces = _oobs_list_new (OOBS_TYPE_IFACE_WIRELESS);
   priv->irlan_ifaces = _oobs_list_new (OOBS_TYPE_IFACE_IRLAN);
@@ -111,6 +246,8 @@ oobs_ifaces_config_init (OobsIfacesConfig *config)
   priv->modem_ifaces = _oobs_list_new (OOBS_TYPE_IFACE_MODEM);
   priv->isdn_ifaces = _oobs_list_new (OOBS_TYPE_IFACE_ISDN);
   config->_priv = priv;
+
+  iface_state_monitor_init (config, oobs_ifaces_config_iface_monitor);
 }
 
 static void
@@ -132,6 +269,12 @@ free_configuration (OobsIfacesConfig *config)
 
   g_list_foreach (priv->available_key_types, (GFunc) g_free, NULL);
   g_list_free (priv->available_key_types);
+
+  if (priv->ifaces)
+    {
+      g_hash_table_destroy (priv->ifaces);
+      priv->ifaces = NULL;
+    }
 }
 
 static void
@@ -150,6 +293,10 @@ oobs_ifaces_config_finalize (GObject *object)
       g_object_unref (priv->plip_ifaces);
       g_object_unref (priv->modem_ifaces);
       g_object_unref (priv->isdn_ifaces);
+
+#ifdef HAVE_HAL
+      libhal_ctx_free (priv->hal_context);
+#endif
     }
 
   if (G_OBJECT_CLASS (oobs_ifaces_config_parent_class)->finalize)
@@ -169,7 +316,8 @@ create_iface_from_message (DBusMessage     *message,
   dbus_message_iter_recurse (iter, &struct_iter);
 
   dev = utils_get_string (&struct_iter);
-  dbus_message_iter_next (&struct_iter);
+  active = utils_get_int (&struct_iter);
+  is_auto = utils_get_int (&struct_iter);
 
   switch (type)
     {
@@ -197,27 +345,17 @@ create_iface_from_message (DBusMessage     *message,
     {
       const gchar *address, *netmask, *gateway, *config_method;
 
-      dbus_message_iter_get_basic (&struct_iter, &is_auto);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &active);
-      dbus_message_iter_next (&struct_iter);
-
       /* This value is deprecated */
       dbus_message_iter_next (&struct_iter);
 
       address = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
-
       netmask = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
 
       /* FIXME: missing network and broadcast */
       dbus_message_iter_next (&struct_iter);
       dbus_message_iter_next (&struct_iter);
 
       gateway = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
 
       g_object_set (iface,
 		    "auto", is_auto,
@@ -232,16 +370,12 @@ create_iface_from_message (DBusMessage     *message,
 	  const gchar *essid, *key, *key_type;
 
 	  essid = utils_get_string (&struct_iter);
-	  dbus_message_iter_next (&struct_iter);
 
 	  /* This value is deprecated */
 	  dbus_message_iter_next (&struct_iter);
 
 	  key = utils_get_string (&struct_iter);
-	  dbus_message_iter_next (&struct_iter);
-
 	  key_type = utils_get_string (&struct_iter);
-	  dbus_message_iter_next (&struct_iter);
 
 	  g_object_set (iface,
 			"essid", essid,
@@ -251,24 +385,14 @@ create_iface_from_message (DBusMessage     *message,
 	}
 
       config_method = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
       g_object_set (iface, "config-method", config_method, NULL);
     }
   else if (OOBS_IS_IFACE_PLIP (iface))
     {
       const gchar *address, *remote_address;
 
-      dbus_message_iter_get_basic (&struct_iter, &is_auto);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &active);
-      dbus_message_iter_next (&struct_iter);
-
       address = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
-
       remote_address = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
 
       g_object_set (iface,
 		    "auto", is_auto,
@@ -282,17 +406,8 @@ create_iface_from_message (DBusMessage     *message,
       const gchar *phone_number, *phone_prefix, *login, *password;
       gboolean default_gw, peer_dns, persistent, noauth;
 
-      dbus_message_iter_get_basic (&struct_iter, &is_auto);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &active);
-      dbus_message_iter_next (&struct_iter);
-
       phone_number = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
-
       phone_prefix = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
 
       if (OOBS_IS_IFACE_MODEM (iface))
 	{
@@ -300,13 +415,8 @@ create_iface_from_message (DBusMessage     *message,
 	  gint volume, dial_type;
 
 	  serial_port = utils_get_string (&struct_iter);
-	  dbus_message_iter_next (&struct_iter);
-
-	  dbus_message_iter_get_basic (&struct_iter, &volume);
-	  dbus_message_iter_next (&struct_iter);
-
-	  dbus_message_iter_get_basic (&struct_iter, &dial_type);
-	  dbus_message_iter_next (&struct_iter);
+	  volume = utils_get_int (&struct_iter);
+	  dial_type = utils_get_int (&struct_iter);
 
 	  g_object_set (iface,
 			"serial-port", serial_port,
@@ -316,22 +426,12 @@ create_iface_from_message (DBusMessage     *message,
 	}
 
       login = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
-
       password = utils_get_string (&struct_iter);
-      dbus_message_iter_next (&struct_iter);
 
-      dbus_message_iter_get_basic (&struct_iter, &default_gw);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &peer_dns);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &persistent);
-      dbus_message_iter_next (&struct_iter);
-
-      dbus_message_iter_get_basic (&struct_iter, &noauth);
-      dbus_message_iter_next (&struct_iter);
+      default_gw = utils_get_int (&struct_iter);
+      peer_dns = utils_get_int (&struct_iter);
+      persistent = utils_get_int (&struct_iter);
+      noauth = utils_get_int (&struct_iter);
 
       g_object_set (iface,
 		    "auto", is_auto,
@@ -355,11 +455,13 @@ static void
 create_ifaces_list (DBusMessage     *reply,
 		    DBusMessageIter *iter,
 		    OobsIfaceType    type,
-		    OobsList        *list)
+		    OobsList        *list,
+		    GHashTable      *ifaces)
 {
   GObject *iface;
   OobsListIter list_iter;
   DBusMessageIter elem_iter;
+  const gchar *name;
 
   dbus_message_iter_recurse (iter, &elem_iter);
 
@@ -369,10 +471,15 @@ create_ifaces_list (DBusMessage     *reply,
 
       oobs_list_append (list, &list_iter);
       oobs_list_set (list, &list_iter, iface);
-      g_object_unref (iface);
 
+      name = oobs_iface_get_device_name (OOBS_IFACE (iface));
+      g_hash_table_insert (ifaces, (gpointer) name, iface);
+
+      g_object_unref (iface);
       dbus_message_iter_next (&elem_iter);
     }
+
+  dbus_message_iter_next (iter);
 }
 
 static void
@@ -388,29 +495,19 @@ oobs_ifaces_config_update (OobsObject *object)
   /* First of all, free the previous configuration */
   free_configuration (OOBS_IFACES_CONFIG (object));
 
+  priv->ifaces = g_hash_table_new (g_str_hash, g_str_equal);
+
   dbus_message_iter_init (reply, &iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_ETHERNET, priv->ethernet_ifaces);
 
-  dbus_message_iter_next (&iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_WIRELESS, priv->wireless_ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_ETHERNET, priv->ethernet_ifaces, priv->ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_WIRELESS, priv->wireless_ifaces, priv->ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_IRLAN, priv->irlan_ifaces, priv->ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_PLIP, priv->plip_ifaces, priv->ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_MODEM, priv->modem_ifaces, priv->ifaces);
+  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_ISDN, priv->isdn_ifaces, priv->ifaces);
 
-  dbus_message_iter_next (&iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_IRLAN, priv->irlan_ifaces);
-
-  dbus_message_iter_next (&iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_PLIP, priv->plip_ifaces);
-
-  dbus_message_iter_next (&iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_MODEM, priv->modem_ifaces);
-
-  dbus_message_iter_next (&iter);
-  create_ifaces_list (reply, &iter, OOBS_IFACE_TYPE_ISDN, priv->isdn_ifaces);
-
-  dbus_message_iter_next (&iter);
-  priv->available_config_methods = utils_get_string_list_from_dbus_reply (reply, iter);
-
-  dbus_message_iter_next (&iter);
-  priv->available_key_types = utils_get_string_list_from_dbus_reply (reply, iter);
+  priv->available_config_methods = utils_get_string_list_from_dbus_reply (reply, &iter);
+  priv->available_key_types = utils_get_string_list_from_dbus_reply (reply, &iter);
 }
 
 static void
@@ -432,13 +529,12 @@ create_dbus_struct_from_iface (DBusMessage     *message,
   dbus_message_iter_open_container (array_iter, DBUS_TYPE_STRUCT, NULL, &iter);
 
   utils_append_string (&iter, dev);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &active);
-  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &is_auto);
+  utils_append_int (&iter, active);
+  utils_append_int (&iter, is_auto);
 
   if (OOBS_IS_IFACE_ETHERNET (iface))
     {
       gchar *address, *netmask, *gateway, *config_method;
-      gint pad = 0;
 
       g_object_get (G_OBJECT (iface),
 		    "ip-address", &address,
@@ -448,7 +544,7 @@ create_dbus_struct_from_iface (DBusMessage     *message,
 		    NULL);
 
       /* This field is deprecated */
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &pad);
+      utils_append_int (&iter, 0);
       utils_append_string (&iter, (configured) ? address : NULL);
       utils_append_string (&iter, (configured) ? netmask : NULL);
 
@@ -471,7 +567,7 @@ create_dbus_struct_from_iface (DBusMessage     *message,
 	  utils_append_string (&iter, (configured) ? essid : NULL);
 
 	  /* This field is deprecated */
-	  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &pad);
+	  utils_append_int (&iter, 0);
 	  utils_append_string (&iter, (configured) ? key : NULL);
 	  utils_append_string (&iter, (configured) ? key_type : NULL);
 
@@ -533,18 +629,18 @@ create_dbus_struct_from_iface (DBusMessage     *message,
 			NULL);
 
 	  utils_append_string (&iter, (configured) ? serial_port : NULL);
-	  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &volume);
-	  dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &dial_type);
+	  utils_append_int (&iter, volume);
+	  utils_append_int (&iter, dial_type);
 
 	  g_free (serial_port);
 	}
 
       utils_append_string (&iter, (configured) ? login : NULL);
       utils_append_string (&iter, (configured) ? password : NULL);
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &default_gw);
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &peer_dns);
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &persistent);
-      dbus_message_iter_append_basic (&iter, DBUS_TYPE_INT32, &noauth);
+      utils_append_int (&iter, default_gw);
+      utils_append_int (&iter, peer_dns);
+      utils_append_int (&iter, persistent);
+      utils_append_int (&iter, noauth);
 
       g_free (phone_number);
       g_free (prefix);
@@ -695,7 +791,6 @@ oobs_ifaces_config_commit (OobsObject *object)
 
 /**
  * oobs_ifaces_config_get:
- * @session: An #OobsSession.
  * 
  * Returns the #OobsIfacesConfig singleton, which represents
  * the network interfaces and their configuration.
@@ -703,23 +798,11 @@ oobs_ifaces_config_commit (OobsObject *object)
  * Return Value: the singleton #OobsIfacesConfig object.
  **/
 OobsObject*
-oobs_ifaces_config_get (OobsSession *session)
+oobs_ifaces_config_get (void)
 {
-  static OobsObject *object = NULL;
-
-  g_return_val_if_fail (session != NULL, NULL);
-  g_return_val_if_fail (OOBS_IS_SESSION (session), NULL);
-
-  if (!object)
-    {
-      object = g_object_new (OOBS_TYPE_IFACES_CONFIG,
-			     "session", session,
-			     "remote-object", IFACES_CONFIG_REMOTE_OBJECT,
-			     NULL);
-      oobs_object_update (object);
-    }
-
-  return object;
+  return g_object_new (OOBS_TYPE_IFACES_CONFIG,
+		       "remote-object", IFACES_CONFIG_REMOTE_OBJECT,
+		       NULL);
 }
 
 /**
